################################################################################
# audioCartography-prepare-generateTrials.R
#
# Title:      Audio Cartography - Prepare Data (Generate Trials)
# Copyright:  2017 University of Oregon
# Author:     M. Brittell
# License:    GNU General Public License, version 3 (GPL3)
# 
# This work was funded in part by the National Science Foundation (NSF) 
# Doctoral Dissertation Research Improvement (DDRI) Grant #1634086 and the 
# University of Oregon (UO) Lewis Family Endowment.
# 
################################################################################
# Systematically generate rasterized data representing spatial distribution of 
# point data synthesized as a point pattern and then aggregated into raster 
# cells by count.  The script produces up to five output files.  Resulting data 
# is stored in three equivalent formats: matrix (2D, nested arrays delimited by 
# square brackets), tabular (one row for each change, for each level of the 
# classed data; only produced for 3-class maps), and a graphic (choropleth map).  
# A summary of descriptive stats includes the number of grid cells assigned to 
# each class.  If the grid size is 8x8, the script attempts to identify 
# candidate response options for trials in the planned experiment (if no 
# feasible options are found in the generated data, no "candidateResponses" 
# file is output).  
# 
# A Python script (audioCartography-prepare-renderTrials.py) renders the audio
# maps.  The choice to use an external script in another language was made out 
# of convenience and familiarity with available audio synthesis functionality 
# in Python via the Pyo library.
################################################################################

# LIBRARIES

# helper function for simplified output when automated
loadLibrary = function(lib){
	cat('   Loading',lib,'... ')
	suppressMessages(library(lib, character.only=TRUE))
	cat('loaded.\n')
}

# handle command line arguments (and parsing)
loadLibrary('docopt') 

# Library to support simulated clustered point patterns
loadLibrary('spatstat')

# Libraries to provide functionality to handle the simple rasterized data
# as spatial data (i.e., using spatial data formats)
loadLibrary('sp')
loadLibrary('rgeos')
loadLibrary('maptools') 

# Libraries to process attribute data: divide into classed intervals
loadLibrary('classInt')

# Support for a simple color ramp
loadLibrary('RColorBrewer')


################################################################################

### PARSE COMMAND LINE ARGUMENTS

doc <- "Usage: generateTrials.r -o OUTDIR -i INDEX [-d]

-o --outdir OUTDIR		Directory in which to write the generated data
-i --index INDEX		Unique identifier used in the output filenames
-d --debug  			Verbose output to help with debugging"

opt <- docopt(doc)

# Optional verbose output
opt$debug <- (opt$debug == 'TRUE')

cat('   Output will be written to:',opt$outdir,'\n')
cat('   Filename identifier: ',opt$index,'\n')
cat('   Verbose output:',opt$debug,'\n')

################################################################################

### CONFIGURATION (CONSTANTS)
# Note: these constants are stated at the top of the script to draw attention 
# to the resolution of the map and symbology that, although flexible, were 
# implemented (and tested) specifically to support evaluation in a specific
# experiment.  The constants could be provided as script parameters, but values 
# other than the intended 8x8 and 3-level will produce a subset of the possible 
# output files.  
#
# This script produces five output files only if GRID_SIZE = 8 (which uses the 
# a known definition of "feasible" response options as 2x2 squares that meet 
# trial criteria for the experiment) and N_LEVELS=3.
#
# Other grid sizes or numbers of data classes can be rendered (output files: 
# matrix, graphic, stats), but will not be accompanied by a list of candidate 
# response options (defined only for 8x8 grid) or tabular representation to 
# support concurrent rendering (implemented only for a three-class map; ).

# -- Grid resolution 
# The grid is always square (number of rows = number of columns; intended to 
# have isometric cells, i.e., rendered in a space where the extent in north-
# south is equal to the extent east-west)
GRID_SIZE = 8
cat('   Raster size: ',GRID_SIZE,'x',GRID_SIZE)
if (GRID_SIZE != 8) cat(' ("feasible" responses only identified for size 8x8)')
cat('\n')

# -- Number of levels in the classed data values 
# In the output, the numeric class identifier is proportional to the density
# of generated points that fall within a grid cell. 
# The maximum number of data classes is 9, limited by the number of colors
# in the avaialble color palette (see below)
N_LEVELS=3
cat('   Number of data classes: ', N_LEVELS)
if (N_LEVELS != 3) cat(' (tabular output only implemented for 3-class maps)')
cat('\n')



################################################################################

### HELPERS

# color ramp for visualizations (maximum number of colors in the palette=9)
blues=brewer.pal(N_LEVELS,"Blues")

# create a spatial (multi-)polygon to represent the GRID_SIZE*GRID_SIZE raster
polys = lapply(1:GRID_SIZE,function(y){
	yTop = y/GRID_SIZE
	yBottom = (y-1)/GRID_SIZE
	inner = lapply(1:GRID_SIZE,function(x){
		xRight = x/GRID_SIZE
		xLeft = (x-1)/GRID_SIZE
		# list the four corners (repeating the first) clockwise
		xCoords=c(xLeft,xLeft,xRight,xRight,xLeft)
		yCoords=c(yBottom,yTop,yTop,yBottom,yBottom)
		id=paste('p',x,y,sep='~')
		Polygons(list(Polygon(cbind(xCoords,yCoords))),id)
	})
	inner
})
polys.sp=SpatialPolygons(unlist(polys))


# recursively visit neighbor cells to identify clusters
#
# args:
# - rowNum  index of the row currently under consideration
# - colNum  index of the column currently under consideration
# - exact   boolean indicating whether or not the check must
#           be an exact match for the currentValue
#
# return:
# data.frame of the grid coordinates that have been identified
# as part of the cluster
#
# environment:
# - readingOrder  matrix of spatial data values
# - currentValue  the value (0:2) of the cluster is being identified
# 					(not implemented for N_LEVELS!=3)
#
# side effects: 
# - alreadyCounted  matrix of cells for bookkeeping; clear this 
#                   as needed from outside the routine, updated 
#                   in the surrounding environment for use across
#                   levels of recursion and multiple clusters

examineCell = function(rowNum,colNum,exact=T){
	if(opt$debug) cat("Examining: ",rowNum,', ',colNum,'(',currentValue,')\n')
	alreadyCounted[rowNum,colNum]<<-TRUE
	acc = data.frame(x=colNum,y=rowNum)
	# look east
	if(colNum<GRID_SIZE 
			&& !(alreadyCounted[rowNum,colNum+1]) 
			&& (readingOrder[rowNum,colNum+1]==currentValue
			|| (!exact && readingOrder[rowNum,colNum+1]>currentValue))){
		if(opt$debug){
			cat("look east, it's value: ",readingOrder[rowNum,colNum+1],'\n')
		}
		acc=rbind(acc,examineCell(rowNum,colNum+1,exact))
	}
	# look south
	if(rowNum<GRID_SIZE 
			&& !(alreadyCounted[rowNum+1,colNum]) 
			&& (readingOrder[rowNum+1,colNum]==currentValue
			|| (!exact && readingOrder[rowNum+1,colNum]>currentValue))){
		if(opt$debug){
			cat("look south, it's value: ",readingOrder[rowNum+1,colNum],'\n')
		}
		acc=rbind(acc,examineCell(rowNum+1,colNum,exact))
	}
	# look west
	if(colNum>1 
			&& !(alreadyCounted[rowNum,colNum-1]) 
			&& (readingOrder[rowNum,colNum-1]==currentValue
			|| (!exact && readingOrder[rowNum,colNum-1]>currentValue))){
		if(opt$debug){
			cat("look west, it's value: ",readingOrder[rowNum,colNum-1],'\n')
		}
		acc=rbind(acc,examineCell(rowNum,colNum-1,exact))
	}
	# look north
	if(rowNum>1 
			&& !(alreadyCounted[rowNum-1,colNum]) 
			&& (readingOrder[rowNum-1,colNum]==currentValue
			|| (!exact && readingOrder[rowNum-1,colNum]>currentValue))){
		if(opt$debug){
			cat("look north, it's value: ",readingOrder[rowNum-1,colNum],'\n')
		}
		acc=rbind(acc,examineCell(rowNum-1,colNum,exact))
	}
	if(opt$debug) print(acc)
	acc
}

# single shot to test the function
#alreadyCounted = matrix(data=FALSE,nrow=GRID_SIZE,ncol=GRID_SIZE)
#currentValue = readingOrder[1,1]
#memberCells = examineCell(1,1)


################################################################################

### GENERATE A DATA SET

cat('   Generating trial ... \n')

# Synthesize data: generate a clustered point pattern
cat('      ... creating point pattern ... ')
values = rMatClust(70,0.2,20)
pts.sp = as.SpatialPoints.ppp(values)
cat('done.\n')

# Rasterize: Rasterize the point data by counting the number of points in each 
# cell, then class the count values into N_LEVELS classes
cat('      ... rasterizing ... ')
dataValues=rowSums(gIntersects(pts.sp,polys.sp,byid=TRUE))
breaks <- classIntervals(dataValues, N_LEVELS, style="quantile")
classedColors=findColours(breaks, blues)
cat('done.\n')

# Data munging: Create alternative data representations
cat('      ... reformatting ... ')
classedValues=as.numeric(factor(
	classedColors,
	levels=attr(classedColors,'palette')
))-1

readingOrder=matrix(
	classedValues,
	nrow=GRID_SIZE,
	byrow=T
)[GRID_SIZE:1,]

# figure out potential response options (which pairs of squares in a 4x4 grid
# both have an "obvious" majority value (three or four of the same value))
if(opt$debug) readingOrder

if(GRID_SIZE==8){
	# "feasible" response options are only defined for GRID_SIZE = 8
	readingOrder.responseOptions = data.frame(
		responseIndex = numeric(),
		dataValue = numeric()
	)
	for (row in 0:3){
		for(col in 1:4){
			responseIndex=row*4+col
			if(opt$debug) print(paste('responseIndex', responseIndex,sep=': '))
			x1=2*(col-1)+1
			x2=2*(col-1)+2
			y1=2*row+1
			y2=2*row+2
			if(opt$debug){
				print(paste('row',paste(c(y1,y2),collapse=','),sep=': '))
				print(paste('column',paste(c(x1,x2),collapse=','),sep=': '))
			}
			v1=readingOrder[y1,x1]
			v2=readingOrder[y1,x2]
			v3=readingOrder[y2,x1]
			v4=readingOrder[y2,x2]
			if(opt$debug) print(rbind(c(v1,v2),c(v3,v4)))
			counts=table(c(v1,v2,v3,v4))
			isValid=any(counts>3)
			if(opt$debug){
				print(paste('value greater than two',isValid,sep="? "))
			}
			if(isValid){
				newIndex=nrow(readingOrder.responseOptions)
				counts=as.data.frame(counts)
				dataValue=counts$Var1[which(counts$Freq>2)]
				readingOrder.responseOptions = rbind(
					readingOrder.responseOptions,
					data.frame(
						responseIndex= responseIndex,
						dataValue= dataValue
					)
				)			
			}
		}
	}
	if(opt$debug) print(readingOrder.responseOptions)
}
	
# SEQUENTIAL: left to right in rows from top to bottom
# put it in reading order and prep for transfer to audio
# marshalling - serialize row-wise
readingOrder.rows=paste(
	'[[',
	paste(
		sapply(1:GRID_SIZE,function(x){
			paste(readingOrder[x,],collapse=",")
		}),
		collapse='],['
	),
	']]',
	sep=''
)
if(opt$debug) readingOrder.rows

# CONTIGUOUS CLUSTERS: Consider 2D extent of contiguous cells with the same 
# class of data).  Contiguous defined as sharing a side (shared corner is not
# considered contiguous); does not consider occlusion (e.g., if a ridge ran
# from the north edge to the south edge, low level on the east side would not 
# be considered "contiguous" with the low level on the west side).

# Initialize accumulator to keep a list of the found clusters
acc=data.frame(
	size=numeric(), value=numeric(), 
	xmin=numeric(),xmax=numeric(),ymin=numeric(),ymax=numeric(),
	centerX=numeric(),centerY=numeric()
)

# Clear the bookkeeping matrix
alreadyCounted = matrix(data=FALSE,nrow=GRID_SIZE,ncol=GRID_SIZE)

# Visit all cells in the raster
for(r in 1:GRID_SIZE){
	for(c in 1:GRID_SIZE){
		if(!alreadyCounted[r,c]){
			currentValue = readingOrder[r,c]
			memberCells=examineCell(r,c)
			temp=data.frame(
				size=nrow(memberCells), value=currentValue,
				xmin=min(memberCells$x), xmax=max(memberCells$x),
				ymin=min(memberCells$y), ymax=max(memberCells$y)
			)
			temp['centerX']=floor(mean(c(temp$xmin,temp$xmax)))
			temp['centerY']=floor(mean(c(temp$ymin,temp$ymax)))
			acc=rbind(acc,temp)
		}
	}
}

# sort by cluster size
readingOrder.contiguous=acc[order(acc$size,decreasing=T),]

# accumulator for summary statistics about the data
# add the known columns first in defining the data frame
readingOrder.stats = data.frame(
	peakCount = length(which(acc$value==2)),
	minClusterSize = min(acc$size),
	avgClusterSize = mean(acc$size),
	maxClusterSize = max(acc$size)
)
# augment with a dynamic number of columns that reflects the 
# number of classes in the classed data.
for (i in 0:(N_LEVELS-1)){
	readingOrder.stats[paste('level',i,sep='')] = sum(classedValues==i)
}
if(opt$debug) print(readingOrder.stats)

# SAMPLED SURFACE: Consider 2D extent of contiguous cells with the same 
# -- or higher -- class of data (makes assumptions about what it means 
# for a higher level to occlude a lower level)
if(N_LEVELS==3){	
	# Initialize accumulator to keep a list of the found clusters
	acc=data.frame(
		value=numeric(), rate=numeric(), frequency=numeric(),
		action=character(),delay=numeric(), duration=numeric()
	)
	
	# PHASE 1: Get the groups of contiguous cells for each level
	# Step through each level (classed attribute data) and identify clusters 
	# of cells with values that are at least as high as that level
	# Skip val=0: the zero value is in all cells (possibly occluded)
	# Note: Can't just do it row by row, the position of the overall group 
	# determines the rate (x coordinate)
	for (currentValue in 1:(N_LEVELS-1)){
		# Clear the bookkeeping matrix
		alreadyCounted = matrix(data=FALSE,nrow=GRID_SIZE,ncol=GRID_SIZE)

		# Visit all cells in the raster
		for(r in 1:GRID_SIZE){		
			for(c in 1:GRID_SIZE){
				if(!alreadyCounted[r,c]){
					if(readingOrder[r,c]<currentValue){
						# mark as visited (ignore in this round)
						alreadyCounted[r,c]=T
					}else{
						# visit this cell (and recurse to its neighbors)
						# allow occlusion (exact=F)
						# result is a cluster (data.frame containing all the 
						# cells that belong to cluster)
						memberCells=examineCell(r,c,F)

						# for each frequency, add a record that indicates:
						# - volume at which to play
						temp=data.frame(value=currentValue)

						# - the rate at which to play (center of cluster)
						if(opt$debug){
							cat('r: ',r,'c: ',c,'memberCells:','\n')
							print(memberCells)
						}
						temp$rate=floor(mean(c(min(memberCells$x),
						max(memberCells$x))))
						if(opt$debug){
							cat('and the calculated rate: ',temp$rate,'\n')
						}
						# - when to turn on, when to turn off (may be multiple 
						#   pairs of on/off)
						onoff = c()

						for(y in unique(memberCells$y)){
							temp$frequency=y

							tempX = sort(memberCells$x[which(memberCells$y==y)])
							start = tempX[1]
							duration = 1

							# value of the cluster with which this action is 
							# associated
							temp$value=currentValue

							if(length(tempX)>1){
								for(x in 2:length(tempX)){
									if(tempX[x]-tempX[x-1]==1){
										# if there is a next one 
										duration = duration+1
									}else{
										# make new record ... 
										# ... on: when it starts
										temp$duration = duration
										temp$action='on'
										temp$value=currentValue
										temp$delay=start-1
										acc=rbind(acc,temp)
										# ... off: when it ends
										temp$duration = NA
										temp$action='off'
										temp$value=-1
										temp$delay=temp$delay+duration
										acc=rbind(acc,temp)

										# and set up for the next record
										start = tempX[x]
										duration = 1
									}
								}
								# make sure the last one gets added
								temp$duration = duration
								# ... turn on
								temp$action='on'
								temp$value=currentValue
								temp$delay=start-1
								acc=rbind(acc,temp)
								temp$duration = NA
								# ... turn off
								temp$action='off'
								temp$value=-1
								temp$delay=temp$delay+duration
								acc=rbind(acc,temp)
							}else{
								# a single cell, make an entry with duration 1
								temp$duration = duration
								# ... turn on
								temp$action='on'
								temp$value=currentValue
								temp$delay=start-1
								acc=rbind(acc,temp)
								# ... turn off
								temp$duration = NA
								temp$action='off'
								temp$value=-1
								temp$delay=start
								acc=rbind(acc,temp)
							} # end else (if tempX)
						} # end for y
					} # end else (if mark visited/ignore)
				} # end if not checked
			} # end for columns
		} # end for rows
	}
	# add an identifier that is unique to each action 
	acc$index=1:nrow(acc)
	# add a copy of the action (later modifying one in place)
	acc$originalAction = acc$action
	# then print it out to see values
	if(opt$debug) acc
	
	# PHASE 2: If symmetric across a vertical axis, level 2 should just 
	# increase volume (value).  Step through each row and handle occlusion 
	# (where level 2 comes on or off and level 1 either acts in the same 
	# way at the same time, or is playing already.)

	toRemove = c()
	for(f in 1:GRID_SIZE){
		# get the list of notes that play in this row (same frequency)
		fi = which(acc$frequency==f)
		# and at the same rate
		for(r in 1:GRID_SIZE){
			fri = intersect(fi,which(acc$rate==r))
			# Process all notes that get played
			if(length(fri)>0){
				# initial sound level
				playingLevel1 = FALSE

				for(d in 0:GRID_SIZE){
					# Step through each delay for which the note changes
					# start at the beginning (no delay) and walk to 
					# the max where all notes should turn off
					dfri = intersect(fri,which(acc$delay==d))

					# if none, no change so no action required
					if(length(dfri)==0) next

					# if one, check to see if it is a level 2 
					# (on:louder or off:softer)
					if(length(dfri)==1){
						if(acc$value[dfri]==2){
							# handle (a single) level 2
							# if only one, then level 1 must already be playing 
							# (i.e., turned on previously, so make louder
							# or should continue playing, but softer)
							if(acc$action[dfri]=='on'){
								acc$action[dfri]='louder'
							}
							if(acc$action[dfri]=='off'){
								acc$action[dfri]='softer'
							}
						}else{
							# handle (a single) level 1
							playingLevel1 = (acc$action[dfri]=='on')
						}
						if(opt$debug) print(paste(r,f,d,'single',sep=", "))
						if(opt$debug) print(acc[dfri,])
					}else if(length(dfri)==2){
						# if more than one, both level 2 and level 1 have the 
						# same action, so remove duplicate
						if(opt$debug) print(paste(r,f,d,'double',sep=", "))
						if(opt$debug) print(acc[dfri,])

						if(all(acc$action[dfri]=='on')){
							playingLevel1 = TRUE
							index1 = intersect(
								acc$index[dfri],
								acc$index[acc$value==1]
							)
							toRemove = c(toRemove,index1)
							if(opt$debug) cat('removing...',index1,'\n')
						}else if (all(acc$action[dfri]=='off')){
							playingLevel1 = FALSE
							toRemove = c(toRemove,dfri[1])
							if(opt$debug) cat('removing...',dfri[1],'\n')
						}
					}else{
						warning("Tabular: not implemented for >3 classes)")
						if(opt$debug) print(paste(r,f,d,'not 1 or 2',sep=", ")) 
						if(opt$debug) print(acc[dfri,])
					} # end cases (number of actions at this position)
				}# end stepping through delay values (min: 0, max:GRID_SIZE)
			}# end check if there are any instances to handle
		}# end for r (rate) - iterate through x dimension
	}# end for f (frequency) - iterate through y dimension
	if(length(toRemove)>0) acc=acc[-toRemove,]
	
	readingOrder.surface=acc
	# sort by cluster centers (optional, more human friendly for debugging)
	readingOrder.surfaceSort=acc[order(acc$delay),]
	readingOrder.surfaceSort=readingOrder.surfaceSort[
		order(readingOrder.surfaceSort$frequency),
	]
	readingOrder.surfaceSort=readingOrder.surfaceSort[
		order(readingOrder.surfaceSort$rate),
	]
}

cat('done.\n')

### summary of what has been prepared:
if(opt$debug){
	# sequential in reading order (left to right and top to bottom; row-wise)
	print("Sequential reading order: ")
	print(readingOrder.rows)

	# groups of contiguous cells that share the same value 
	# (does not consider occlusion)
	print("Contiguous groups: ")
	print(readingOrder.contiguous)

	# groups of contiguous cells that share the same minimum value 
	# (occlusion of lower values by higher values; only if N_LEVELS==3)
	if(N_LEVELS==3){
		print("Surface representation: ")
		print(readingOrder.surface)

		print("Surface representation... sorted: ")
		# same as readingOrder.surface but with optional (human friendly) sort
		print(readingOrder.surfaceSort)
	}
	# The result (if N_LEVELS==3):
	# One entry in the output for each group of contiguous cells that have the 
	# same or a higher class of data ("or higher" captures concept of 
	# occlusion: if there is level 3 data, then it is at least level 2 and gets 
	# represented with both level 2 and level 3 symbols).
}

cat('   Writing output ... \n')

# write output to the location specified as an input parameter
setwd(opt$outdir)

# save the trials that have candidate response options 
# NOTE: "Feasible" response options are only defined for 8x8 grid
if(GRID_SIZE==8 && nrow(readingOrder.responseOptions)>1){
	# save a list of the candidate response indices and data values
	cat('      ... candidate response options ... ')
	write.csv(
		readingOrder.responseOptions,
		paste("candidateResponses-",opt$index,".csv",sep=''),
		row.names=F
	)
	cat('done.\n')
}

# save the trials in four formats (graphic, matrix - for sequential, 
# tabular - for concurrent, descriptive statistics)

# save the visual graphic to a file
cat('      ... graphic ... ')
pdf(paste("graphic-",opt$index,".pdf",sep=''))
plot(polys.sp,border=NA,col=classedColors)
invisible(dev.off())
cat('done.\n')

# save the sequential data by row
cat('      ... sequential ... ')
write(readingOrder.rows,paste("sequential-",opt$index,".txt",sep=''))
cat('done.\n')

# note concurrent playback implementation limited to 3 data levels
if(N_LEVELS==3){
	# save the concurrent representation of data (with occlusion); subset
	# to only the columns required for rendering (plus an identifier: index)
	cat('      ... concurrent ... ')
	write.csv(
		readingOrder.surface[c('index','rate','frequency','value','delay')],
		paste("concurrent-",opt$index,".csv",sep=''),
		row.names=F,quote=F
	)
	cat('done.\n')
}

# save descriptive statistics about the generated data
cat('      ... stats ... ')
write.csv(
	readingOrder.stats,
	paste("stats-",opt$index,".csv",sep=''),
	row.names=F
)
cat('done.\n')

cat('Script complete.\n\n')
