#!/bin/bash

################################################################################
# audioCartography-prepare-makeMaps.sh
#
# Title:      Audio Cartography - Prepare Data  (Make Maps)
# Copyright:  2017 University of Oregon
# Author:     M. Brittell
# License:    GNU General Public License, version 3 (GPL3)
# 
# This work was funded in part by the National Science Foundation (NSF) 
# Doctoral Dissertation Research Improvement (DDRI) Grant #1634086 and the 
# University of Oregon (UO) Lewis Family Endowment.
# 
################################################################################
# Automate the generation and rendering of multiple instances of audio maps 
# using three audio map types: sequential, augmented-sequential (hybrid), and 
# concurrent.
################################################################################

# Usage, check for a single, numeric argument
if [[ $# != 1 || $1 =~ [^0-9] ]]; then 
	echo $'\n'"USAGE: audioCartography-prepare-makeMaps.sh <number>"$'\n'
	echo "A single numeric argument must specify how many maps to make."$'\n'
	exit
else
	# Number of maps to generate, stored as a new variable to indicate
	# that it passed the check for being a number.
	N=$1
	if [[ $N > 1 ]]; then plural='s'; fi
	echo "--- Generating $N map$plural ... "
fi

# Location of the data preparation scripts
# If scripts are located in different folders, replace with a string that 
# specifies the folder containing the scripts that do the work of generating 
# data sets and rendering those data
SCRIPTDIR="."

# Directory where data files should be written
# Location is relative to the working directory from which the script is called 
# or an absolute path may be specified; data files include the matrix and 
# tabular data formats as well as a graphic for visual inspection
DATADIR="_data"
mkdir -p $DATADIR
echo '--- Matrix and tabular files will be written to:'$'\n\t'$DATADIR

# Directory where the rendered audio files should be written
# Location is relative to the working directory from which the script is called 
# or an absolute path may be specified; three audio files are rendered for each 
# generated data set
AUDIODIR="_audio"
mkdir -p $AUDIODIR
echo '--- Rendered audio files will be written to:'$'\n\t'$AUDIODIR


# Generate maps
for i in `seq 1 $N`;
do
	# Call R code to render an instance of an 8x8 raster with 3-level classed 
	# data that is represented in, five output files:
	# - "sequential-<N>" sequential data representation (nested arrays)
	# - "concurrent-<N>" concurrent data representation (list of audio changes)
	# - "graphic-<N>" visualization (for troubleshooting)
	# - "stats-<N>" descriptive statistics (description of data properties)
	# - "viable-<N>" candidate response option locations, if any (16 possible) 
	echo "    ... Generating instance $i"
	Rscript $SCRIPTDIR/audioCartography-prepare-generateTrials.R \
		-o $DATADIR -i $i

	# Call Python code to render in audio across all three conditions
	# Two versions of the sequential rendering: 
	# - "sequential" uses constant frequency and note durations; 
	# - "sequential2" varies frequency and note duration.
	# One version of conurrent rendering (with mulitple repetitions)
	echo "   ... Rendering audio for instance $i"
	# File naming of the data (inputs) matches those output from 
	# audioCartography-prepare-generateTrials.R
	python $SCRIPTDIR/audioCartography-prepare-renderTrials.py \
		-t sequential \
		-o "$AUDIODIR/sequential-$i.wav" \
		-d "$DATADIR/sequential-$i.txt"	
	python $SCRIPTDIR/audioCartography-prepare-renderTrials.py \
		-t sequential2 \
		-o "$AUDIODIR/hybrid-$i.wav" \
		-d "$DATADIR/sequential-$i.txt"
	python $SCRIPTDIR/audioCartography-prepare-renderTrials.py \
		-t concurrent \
		-o "$AUDIODIR/concurrent-$i.wav" \
		-d "$DATADIR/concurrent-$i.csv"

done

echo "Done."
