################################################################################
# audioCartography-prepare-renderTrials.py
#
# Title: 		Audio Cartography - Prepare Data  (Render Audio)
# Copyright: 	2017 University of Oregon
# Author: 		M. Brittell
# License: 		GNU General Public License, version 3 (GPL3)
# 
# This work was funded in part by the National Science Foundation (NSF) 
# Doctoral Dissertation Research Improvement (DDRI) Grant #1634086 and the 
# University of Oregon (UO) Lewis Family Endowment.
# 
################################################################################
# Generate auditory representations of spatial data that has been encoded as 
# either a two dimensional array of values (sequential and sequential2) or a 
# table of audio synthesis actions (concurrent).
# 
# An R script (audioCartography-prepare-generateTrials.R) prepares data for 
# rendering.  The choice to use an external script in another language was made 
# out of convenience and familiarity with available spatial data processing 
# functionality in R.
################################################################################

################################################################################
###                           EXAMPLE COMMAND LINE                           ###
################################################################################

"""
### Sequential map (no output specified: audio out)
python audioCartography-prepare-renderTrials.py \
	-t sequential \
	-d ../audioCartography-maps/audioCartography-maps-matrix/data01.txt

### Sequential map (outfile specified: render offline, write to file)
python audioCartography-prepare-renderTrials.py \
	-o sequential01.wav\
	-t sequential \
	-d ../audioCartography-maps/audioCartography-maps-matrix/data01.txt

### Augmented sequential map 
python audioCartography-prepare-renderTrials.py \
	-o augmented01.wav\
	-t sequential2 \
	-d ../audioCartography-maps/audioCartography-maps-matrix/data01.txt

### Concurrent map 
python audioCartography-prepare-renderTrials.py \
	-o concurrent01.wav \
	-t concurrent \
	-d ../audioCartography-maps/audioCartography-maps-tabular/table01.csv
"""

################################################################################
###                                LIBRARIES                                 ###
################################################################################
# accept and handle command line arguments that modify the default configuration
import argparse

# audio library that does the rendering and handles writing audio files
from pyo import *

# for exponent computation of attenuation values
import math

# to interpret data as entered in an array of arrays (sequential, sequential2)
from numpy import array

# to read a table with note playback actions (concurrent)
import csv


################################################################################
###                                CONSTANTS                                 ###
################################################################################

# Time required to acquire one image volume in seconds
TA = 2
"""
The time required to acquire a single volume of fMRI data in a multiband 
sequence.  The rendered audio output includes intentional delays to synchronize
(offset) the auditory stimulus with the acquisition noise of the scanner; 
incorrect values or inaccuracies in duration will accumulate over the duration 
of the trial (slipping).
"""

# Time between image acquistion (end to start) in seconds
SILENT_INTERVAL = 5
"""
The duration of the silent interval in a sparse sampling fMRI sequence.  
The audio output includes rendered stimulus sounds only during the silent 
interval, interleaved with the acquisition sequence.  The usable stimulus 
duration is less than the silent interval

see also:
	SILENCE			The SILENT_INTERVAL must be greater than twice the SILENCE
"""

# Period between the stimulus and the acquisition in seconds
SILENCE = 0.5
"""
The silent duration is inserted both before and after the stimulus; when 
computing the overall duration, it should be doubled.  The silence falls within 
the silent interval 

see also:
	SILENT_INTERVAL	SILENCE must be less than half the SILENT_INTERVAL
"""

# Map resolution: spatial data rasterized/discretized in 8x8 grid
GRID_SIZE = 8
"""
This value is used to control the number of loop iterations; values smaller 
than 8 could be used without further modification to this script, larger values
would require additional values for the frequency and duration variables.  
The audio rendering would use a subset of frequencies and note durations and 
rates; the subset is the first GRID_SIZE values in the respective value lists 
and utilizes only part of the available dynamic range, i.e., it does NOT try to 
distribute values across the available range to maximize differences between 
values.

see also: 
	FREQUENCY	GRID_SIZE must be less than or equal to the number of 
				frequencies defined in FREQUENCY
"""

# Amplitude (volume), which encodes data values
# Correction 1 of 3: remove the values over range (0,1) below that overwrite
# the values over range (0,100)
# volume levels 0 to 100 (adjusted to achive audible levels)
VOLUME = [10,40,85]
# volume levels 0 to 1 (as used for the study)
VOLUME = [0.10,0.40,0.85]
"""
Volume values in three discrete levels to represent data value (magnitude).  
The maximum value is less than one to reduce the occurence of clipping in 
the rendered output.  The minimum value was chosen to be small, but still 
audible.  The middle value offered a subjective balance between the endpoints
(range 0 to 100).
""" 

# Note durations within symbols (fraction between 0 and 1)
DURATION = [1.0/x for x in range(1,GRID_SIZE+1)]
"""
The auditory symbols are specified by numeric values that describe the physical 
properties of sound waves.  Duration (rate) was chosen to represent longitudes
and stored in an (ordered) array of fractional values.

see also: 
	GRID_SIZE		length(DURATION) is GRID_SIZE
"""

# Index into DURATION that holds the duration of fixed-duration notes
FIXED_DURATION_INDEX = 0
"""
When rendering the "sequential" auditory map, use a single duration across all 
note symbols.  The temporal order alone encodes location within the two 
dimensional display.

see also: 
	DURATION		list of durations for variable length notes
"""

# Note frequencies within the C major arpeggio
FREQUENCY = {
	'C3':130.813, 'E3':164.814, 'G3':195.998,
	'C4':261.626, 'E4':329.628, 'G4':391.995,
	'C5':523.251, 'E5':659.255
}
"""
The auditory symbols are specified by numeric values that describe the physical 
properties of sound waves.  Frequency (pitch) represents latitudes. 
Frequencies are discretized with a resolution of GRID_SIZE units (display is 
divided into GRID_SIZE=8 regions of equal latitude).

The frequencies were chosen from the C major scale (arpreggio) to avoid 
dissonant chords when multiple notes are played simultaneously.  

Frequencies are stored in a dictionary to be able to refer to them by note 
name.  The values are based on relative difference from A4=440 (A0=27.5 and 
each step an increase by a factor of 2^(1/12)).

see also:
	GRID_SIZE		the resolution of the raster(-ized) display
	DBSPL			dB levels at 20 phon 
"""

# Single frequency used in "sequential" rendering
FIXED_FREQUENCY = 'C4'
"""
When rendering the "sequential" auditory map, use a single frequency.  The
temporal order alone encodes location within the two dimensional display.

see also: 
	FREQUENCY		list of note names (keys) and frequencies (values)
"""

# dB SPL along the 20 phon curve for respective frequencies
# Correction 2 of 3: remove the 40 phon values below that overwrite
# this 20 phon definition
# 20 phon (corrected to match 20 phon used for reference FULL_DB)
DBSPL = {
	'C3':40.0, 'E3':36.0, 'G3':34.0,
	'C4':30.0, 'E4':27.0, 'G4':25.0, 
	'C5':23.0, 'E5':21.0 
}
# 40 phon (as rendered for study)
DBSPL = {
	'C3':57.0, 'E3':53.0, 'G3':51.0,
	'C4':47.0, 'E4':44.0, 'G4':43.0, 
	'C5':41.0, 'E5':40.0
}
"""
The frequencies are attenuated to approximate equal loudness across the range 
of frequencies used in the display. 

Attenuation based on equal loudness contours ISO 226:2003 (as interpreted from
svg by Lindosland at English Wikipedia, 2005, public domain): estimated to 
whole numbers, includes a decimal point to make float type.

see also: 
	FREQUENCY		list of note names (keys) and frequencies (values)
	ATTENUATION		scalars to attenuate rendering based on equal loudness curves
"""

# Reference dB value 
FULL_DB = 90
"""
Frequency based attenuation is determined relative to 20 Hz at 20 phon
"""

# Attenuation scalar relative to FULL_DB
ATTENUATION = {i:math.pow(10,(DBSPL[i]-FULL_DB)/20) for i in DBSPL}
"""
Convert the dB values along the equal loudness contour into scalar values
that are used to control amplitude of the playback.  Physics equation as 
described by Sengpiel (2014).

Amount of attenuation (with sign): DBSPL[i]-FULL_DB
Attenuation = (DBSPL[i]-FULL_DB) = 20 log_10 (output/input)
"""

# The number of times the concurrent map repeats
REPEATS = GRID_SIZE
"""
The 'concurrent' rendering takes substantially less time, the entire playback 
fits within the silent period.  Multiple repetitions allows the listener to 
direct their attention to different aspects of the display across multiple 
instances.  Repetitions also makes the two conditions (sequential and 
concurrent) more comparable in terms of their overall playback time, and the 
number of fMRI volumes collected while listening to each map.
"""

# Attack, Decay, Sustain: description of the temporal envelope of notes
ATTACK_RATIO = 0.35		
"""
The proportion of the time allocated for note playback that is spent on the 
attack for notes that are part of the augmented sequential (sequential2) and 
concurrent map types.  A longer attack ratio (relative to the fixed duration 
notes) gives a more gentle onset for smoother rhythm.
"""
ATTACK_RATIO_F = 0.15	
"""
The proportion of the time allocated for note playback that is spent on attack
for notes that are part of the sequential map, which uses fixed duration notes: 
one note per grid cell.  The sharper onset (relative to the notes the compose 
multi-note map symbols) gives more clear distinction between notes.
"""
DECAY_RATIO = 0.15		 
"""
The proportion of the time allocated for note playback that is spent in decay.
"""
SUSTAIN_RATIO = 0.6 	
"""
The proportion of the time allocated for note playback that is spent on the 
sustain.  Smaller values give larger gaps between repetitions of notes within
the multi-note map symbols.

The total (attack + sustain + decay) may be greater than one, in which case 
there is overlap between repetitions; one note may still be sounding when a
subsequent note begins.
"""


################################################################################
###                             COMPUTED VALUES                              ###
################################################################################

# Temporal separation between onset of audio bursts in seconds
singleShotDuration = SILENT_INTERVAL+TA
"""
Each audio burst plays during a period of relative silence from the scanner.  
The difference between the onset of one audio burst and the next comprises the
amount of time that the scanner is "silent" and the amount of time that it takes
the scanner to acquire a single image volume (and during which the scanner 
produces loud noise).
"""

# Temporal duration of an auditory burst in seconds
mapDuration = SILENT_INTERVAL-(2*SILENCE)
"""
The map duration is the amount of time allocated to the presentation of 
information through audio.  This duration fits within the relative silence 
between image volume acquisitions, and is isolated from direct sound 
interference by a brief silence between the scanner noise and the audio 
burst, and between the audio burst and the subsequent scanner noise.

See also: 
	SILENT_INTERVAL		duration of silence in sparse sampling sequence
	SILENCE				quiet period between stimulus and EPI acquisition
"""

# Maximum duration of a single raster cell in seconds
cellDuration = mapDuration/GRID_SIZE
"""
The maximum temporal duration of a single cell within the GRID_SIZE matrix, 
based on the total available play time (constrained by the sparse sampling 
sequence silent period).
"""

# Parameterized audio symbols used as auditory map symbols
symbols = {}
"""
Accumulator to hold the configuration for auditory symbols; dimensionality 
matches the dimensions of the display (GRID_SIZE x GRID_SIZE).  The values
within the accumulator gets populated later based on the configuration options.
Note: indexed by discrete latitude (Y), longitude (X)
"""


################################################################################
###                          DEFAULT CONFIGURATION                           ###
################################################################################

# Configuration: playback configuration using default values
config = {
	'type':'sequential',
	'out':None,
	'duration':None,
	'data':None,
	'verbose':False,
	'isOffline':False
}
"""
Configuration settings to determine which type (experimental conditions or 
sampler of all auditory symbols).  By default, the script will render the 
"sequential" map type and present it through the system audio (speakers)
"""


################################################################################
###                          COMMAND LINE ARGUMENTS                          ###
################################################################################
""" 
On calling the script, provide information about what type of auditory map to 
generate (sequential, sequential2, concurrent), and where to write the result.  
The extension provided with the out filename determines the format of the output 
file (as implemented in Pyo).

Note: the "sequential2" type is a sequential playback that augments each note 
with location information (encoding location in frequency and note rate in 
addition to temporal order)
"""

# a parser object to extract arguments
parser = argparse.ArgumentParser()

# type (temporal configuration) of audio stream to synthesize
typeHelp = "Playback type [ sequential | sequential2 | concurrent ]; "
typeHelp += "default value: sequential." 
parser.add_argument("-t","--type",help=typeHelp)

# output file (or play over speakers if None)
outHelp = "Output file to which to write synthesized audio.  "
outHelp += "The file extension (e.g., '.wav') determines file format; "
outHelp += "render OGG if unknown extention provided, or "
outHelp += "render to audio out if no -out parameter provided.  "
parser.add_argument("-o","--out", help=outHelp)

# geospatial data represented as a GIRD_SIZE by GIRD_SIZE matrix
dataHelp = "Name of a file from which to read formatted data given as either "
dataHelp += "{SEQUENTIAL[2]:} a comma separated and square bracket delimited "
dataHelp += "nested lists of numeric values (i.e., an array of arrays, e.g., "
dataHelp += "'[[1,2],[3,4]]' with maximum dimensions: "+str(GRID_SIZE)+" by "
dataHelp += str(GRID_SIZE)+"), or "
dataHelp += "{CONCURRENT:} tabular form (CSV) with a row for each action and "
dataHelp += "at least the following columns: setValue, rate, frequency, delay. "
parser.add_argument("-d","--data", help=dataHelp, required=True)

# Debug option to adjust verbosity
verboseHelp = "Flag (no value) used to set the verbosity of script functions"
parser.add_argument("-v","--verbose", help=verboseHelp, action="store_true")

# parse command line arguments, recording in the configuration accumulator
# and setting up initial audio server state
args = parser.parse_args()

# Read flag indicating whether or not to provide verbose output
config['verbose']=args.verbose

# Determine the type of audio symbology (sequential | sequential2 | concurrent)
# default is a sampler enumerating all audio symbols; update the total duration
# as needed
# Note: audio display type must be processed before initializing the 
# 		server to correctly determine total duration of the output file
if args.type: config['type'] = args.type
# otherwise use the default (sequential)
if config['type']=='sequential' or config['type']=='sequential2':
	config['duration'] = singleShotDuration * GRID_SIZE 
	if config['verbose']: print "Audio type: "+config['type']
elif config['type']=='concurrent':
	config['duration'] = singleShotDuration * REPEATS
	if config['verbose']: print "Audio type: "+config['type']
else:
	raise ValueError('Invalid type ("'+config['type']+'")')
	
# provide verbose output for debugging: echo total rendered duration
if config['verbose']: 
	print "Total rendered audio duration: "+str(config['duration'])+' seconds'

# Read and format data given as a string in the command line
if not args.data:
	raise ValueError('No data to render (provide the name of a data file)')
		
# open the specified input file
infile = open(args.data, 'r')

# read sequential representation of data and create a 2D array 
if config['type'].startswith('sequential'):
	d = infile.readline().lstrip('"').rstrip('"\n')
	try:
		config['data']=array(eval(d))
		size = config['data'].shape
	except:
		raise ValueError('Incorrect data format for "sequential"')
	if config['verbose']: 
		print 'Data size: '+str(config['data'].shape)
	if size[0]!=GRID_SIZE or size[1]!=GRID_SIZE:
		raise ValueError('Data is not expected size.')
		
# read concurrent representation of data and create action list
if config['type']=='concurrent':
	config['data'] = list(csv.DictReader(infile))

# provide verbose output for debugging: echo data read from input file
if config['verbose']: print config['data']		

# Handle an output file, if specified, and initialize audio server accordingly
if args.out:
	config['isOffline']=True
	config['out'] = args.out
	# start an audio server to send synthesized waveform to a file
	pyoServer = Server(audio="offline", nchnls=2, duplex=0)
	pyoServer.recordOptions( \
		dur=config['duration'], \
		filename=config['out'], \
		fileformat=0)
	if config['verbose']:
		pyoServer.setVerbosity(8)
	# provide verbose output for debugging: outfile name
  	if config['verbose']: print "Outfile: "+config['out']  
else:
	# start an audio server to send synthesized waveform to audio out as the 
	# code renders the waveform
	pyoServer = Server()

pyoServer.boot()


################################################################################
###                      PYO SERVER DEPENDENT CONSTANTs                      ###
################################################################################

# Woodwind-like timbre for all notes
CLARINET = HarmTable([0.7,0.2,0.8,0.2,0.7,0.2,0.6,0.15,0.5,0.15,0.4,0.1])
"""
Pyo server dependent object for a woodwind-like note with alternating strong 
and weak harmonics.

The table of harmonics must be defined after a Pyo server has been instantiated
"""

################################################################################
###                             HELPER FUNCTIONS                             ###
################################################################################

# Helper function to create object specific functions to modify state
def setVolumeGenerator(f,fader):
	"""	
	Parameters:
		f: `float`	
			frequency of the note, which has a respective equal loudness metric 
		fader: `Fader`	
			Fader object that the resulting function should modify
			
	Returns:
		function that will convert the data value level into a factor and apply
		the result to the corresponding Fader; the result is zero if the note 
		should be turned off
	"""
	# scalar to apply based on the frequency (fixed across value levels)
	eqLoudness = ATTENUATION[f]
	def setVolume(value):
		# the code -1 indicates turning "off", so amplitude = 0
		if value==-1: newVolume=0
		# otherwise look up the factor to apply based on symbology
		else: newVolume = VOLUME[value]
		
		# Correction 3 of 3: remove scaling by a factor of 8
		# prior to the correction for misalignment in the phon level of 
		# the refrence and the note dB, the "newVolume" was adjusted by a
		# factor of 8 so that notes could be heard in the renered stream
		newVolume *= 8
		
		# adjust based on map type: the concurrent map plays multiple
		# notes concurrently so reduces amplitude of all equally to 
		# reduce the occurrance of clipping
		if config['type']=='concurrent': newVolume /= 2
		
		# apply the attenuation to the given note by setting the 
		# multiplier on the Fader
		fader.mul = eqLoudness * newVolume
		
		# feedback for debugging
		if config['verbose']:
			print 'volume value: '+str(value) + \
			', eqLoudness: '+str(eqLoudness) + \
			', mul: '+str(fader.mul)
			
	return setVolume

# Helper function to render audio within a repeated pattern
def playNoteGenerator(fader):
	"""	
	Parameters:
		fader: `Fader`	
			Fader object that the resulting function should modify
			
	Returns:
		function that will play sound (as configured within the Fader)
	"""
	def playNote():
		fader.play()
	return playNote

# CallAfter objects must be saved, used only for offline rendering
# (in unique storage space; don't let them overwrite one another)
adjustments = {}
# Helper method to play individual cells for either online or offline rendering
def playCell(x,y,v,rep=1,delay=0):
	"""
	Parameters:
		x: `int`
			longitude value (discrete valued in range 0:GRID_SIZE-1)
		y: `int`
			latitude value (nominal values in FREQUENCY)
		v: `int`	
			data value (positive numeric value, or -1 to silence)
 		rep: `int`
			repetitions, or duration in number of cells; -1 indicates that the
			playback shoud continue indefinitely
		delay: `float`
			time to wait before playing the note (only used when rendering 
			offline)

	Returns:
		None (side effect of rendering audio)
	"""
	# Note: "symbols" indexed by lat (Y), lon (X)
	if config['isOffline']:
		adjustments[len(adjustments)] = \
			CallAfter(symbols[y][x]['setData'], delay*cellDuration, v)
		if rep>0:
			adjustments[len(adjustments)] = \
			CallAfter(symbols[y][x]['setData'], (delay+rep)*cellDuration, -1)
		if config['verbose']:
			print 'start: '+str(delay*cellDuration)
			print ' end: '+str((delay+rep)*cellDuration)
	else:
		symbols[y][x]['setData'](v)
		# sleep while it plays, then silence the note
		if rep>0:
			time.sleep(rep*cellDuration)
		else:
			print "WARNING: each cell of online playback has fixed duration"
			time.sleep(cellDuration)
		symbols[y][x]['setData'](-1)

################################################################################
###                             CONFIGURE NOTES                              ###
################################################################################

# Instantiate Pyo objects for symbols at each cell position
# store list of candidate values in the environment, parameterized by the
# configuration options.  
# 	durations	length of notes to repeat (in seconds)
#	frequencies	ordered (low to high) tuples of note name, frequency (in Hz)
if config['type'] == 'sequential':
	# Hold the frequency and note duration constant
	durations = [DURATION[FIXED_DURATION_INDEX]]
	frequencies = {FIXED_FREQUENCY:FREQUENCY[FIXED_FREQUENCY]}.items()	
else:
	# location in the display is encoded in frequency and note durations
	durations = DURATION
	# make sure the frequencies are sorted by value (Hz) from low to high
	frequencies = sorted(FREQUENCY.items(), key = lambda(k, v): v, reverse=True)

# provide verbose output for debugging: echo parameters to initialize notes
if config['verbose']:
	print 'the durations: ', durations
	print 'the frequencies: ', frequencies

# Results of the anonymous lambda function contain two elements
# the frequency name (note name) at index zero, and hertz at index one
FNAME = 0; FHZ = 1
# Iterate through all the frequencies
for frequency in frequencies:
	frequencyName = frequency[FNAME]
	# create an entry for each frequency (organize by Y axis first)
	symbols[frequencyName] = {}	
	# Iterate through all the durations (X axis)
	for i,duration in enumerate(durations):
		# create an entry for each duration (within each frequency)
		symbols[frequencyName][i] = {}
		if config['verbose']: print '...Initializing ',frequencyName,',',i
		
		### Initialize pattern components 
		# Note: for Python (Pyo) objects, no reuse across multiple symbols
		#       so instantiate one for each note and store in the accumulator
		#       "symbols" that was declared above and indexed by lat,lon (Y,X)
		# duration of the (possibly repeated) note	
		noteDuration = cellDuration*duration
		symbols[frequencyName][i]['noteDuration'] = noteDuration
		if config['verbose']: 
			print 'duration: ',noteDuration,'(',cellDuration,',',duration,')'	
		
		# temporal envelope of the note
		attackRatio = ATTACK_RATIO
		if config['type']=='sequential': attackRatio = ATTACK_RATIO_F
		symbolNote = Fader( \
			fadein=noteDuration*attackRatio, \
			fadeout=noteDuration*DECAY_RATIO, \
			dur=noteDuration*SUSTAIN_RATIO, \
			mul=1 # no initial attenuation for equal loudness
		) 
		symbols[frequencyName][i]['note'] = symbolNote
		
		# spectral envelope (with given temporal envelope, frequency)
		symbols[frequencyName][i]['timbre'] = \
			Osc(CLARINET, \
				freq=frequency[FHZ], \
				mul=symbolNote).mix(2).out()
			
		# dynamically adjust the amplitude (loudness) based on data value
		symbols[frequencyName][i]['setData'] = \
			setVolumeGenerator(frequencyName,symbolNote)
		
		# function to render the note to the audio stream
		symbols[frequencyName][i]['playNote'] = \
			playNoteGenerator(symbolNote)
		
		# regular repetition
		symbols[frequencyName][i]['pattern'] = \
			Pattern( \
				symbols[frequencyName][i]['playNote'], \
				symbols[frequencyName][i]['noteDuration'])
		
		# Start with the note silent (supports flexibility for online or 
		# offline rendering and "playing" the same symbol multiple times in a 
		# single output stream)
		symbols[frequencyName][i]['setData'](-1)
		
		# set playback duration for each note and start rendering
		if config['isOffline']:
			# let them play (possibly silently) a bit longer than file duration
			paddedDuration = config['duration']+1
			symbols[frequencyName][i]['pattern'].play(dur=paddedDuration)
		else:
			# play indefinitely (while changes are applied to the audio stream)
			symbols[frequencyName][i]['pattern'].play()


################################################################################
###                                RENDERING                                 ###
################################################################################

# When rendering online, start the server before playing any notes
if not config['isOffline']: pyoServer.start()


################################################################################
###                          SYNTHESIZE AUDIO STREAM                         ###
################################################################################

### MAPS
# SEQUENTIAL: 	sequential representation; fixed frequency and duration
if config['type'] == 'sequential':
	for row in range(GRID_SIZE):
		for col in range(GRID_SIZE):
			v = config['data'][row][col]
			delay = row*singleShotDuration+col*cellDuration
			delay /= cellDuration
			playCell(FIXED_DURATION_INDEX,FIXED_FREQUENCY,v,rep=1,delay=delay)
		# pause online rendering for acquisition (just to reflect actual 
		# playback timing; audio will be pre-rendered for scanning sessions)
		if not config['isOffline']: time.sleep(TA+2*SILENCE)

# SEQUENTIAL2: 	sequential representation; varying frequency and duration
if config['type'] == 'sequential2':
	for row in range(GRID_SIZE):
		for col in range(GRID_SIZE):
			f = frequencies[row][FNAME]
			v = config['data'][row][col]
			delay = row*singleShotDuration+col*cellDuration
			delay /= cellDuration
			if config['verbose']: 
				print col, f, v, row*GRID_SIZE+col, delay
			playCell(col,f,v,rep=1,delay=delay)
		# pause online rendering for acquisition (just to reflect actual 
		# playback timing; audio will be pre-rendered for scanning sessions)
		if not config['isOffline']: time.sleep(TA+2*SILENCE)

# CONCURRENT: 	concurrent representation; varying frequency and duration 
if config['type'] == 'concurrent':
	# include a delay for the silence and acquisition between repetitions of 
	# the concurrent representation
	for repeatDelay in [singleShotDuration*i for i in range(REPEATS)]:	
		# repond to each action (row) in the data
		for row in config['data']:
			x = int(row['rate'])-1
			y = frequencies[int(row['frequency'])-1][FNAME]
			v = int(row['value'])
			# amount of time to wait before enacting (in number of cells)
			delay = int(row['delay'])+(repeatDelay/cellDuration)
			# modify the playback (negative reps: explicit on and off actions)
			playCell(x,y,v,rep=-1,delay=delay)
			if config['verbose']: print 'Data row: ',delay,x,y,v


################################################################################
###                                  CLEANUP                                 ###
################################################################################

### Flush buffers and cleanup
if config['isOffline']: 
	# When rendering offline, start after calling the play function(s)
	pyoServer.start()
	exit()
else:
	# when the "playCell" function returns, it is ready for cleanup
	time.sleep(1);pyoServer.stop();time.sleep(0.25);pyoServer.shutdown()
	exit()

################################################################################
